#![feature(rustc_private)]
#![feature(fn_traits)]

#[cfg(feature = "ssr")]
fn setup_tracing() {
    use tracing::level_filters::LevelFilter;
    use tracing_subscriber::EnvFilter;

    cfg_if::cfg_if! {
        if #[cfg(debug_assertions)] {
            let env_filter = EnvFilter::builder()
                .with_default_directive(LevelFilter::TRACE.into())
                // .parse_lossy("debug,llama_forge_rs=debug");
                .parse_lossy("info,llama_forge_rs=debug");
        } else {
            let env_filter = EnvFilter::builder()
                .with_default_directive(LevelFilter::INFO.into())
                .parse_lossy("warn,llama_forge_rs=info");
        }
    }

    tracing_subscriber::fmt()
        .pretty()
        .with_env_filter(env_filter)
        .init();
}

#[cfg(feature = "ssr")]
fn wry_main() -> wry::Result<()> {
    use tao::{
        event::{Event, WindowEvent},
        event_loop::{ControlFlow, EventLoopBuilder},
        platform::unix::EventLoopBuilderExtUnix,
        window::WindowBuilder,
    };
    use wry::WebViewBuilder;

    tracing::debug!("Hello from wry main");

    let event_loop = EventLoopBuilder::new().with_any_thread(true).build();
    let window = WindowBuilder::new()
        .with_min_inner_size(tao::dpi::LogicalSize::new(800, 600))
        .with_title("Llama Forge RS")
        .build(&event_loop)
        .unwrap();

    let builder = {
        use tao::platform::unix::WindowExtUnix;
        use wry::WebViewBuilderExtUnix;
        let vbox = window.default_vbox().unwrap();
        WebViewBuilder::new_gtk(vbox)
    };

    let _webview = builder.with_url("http://localhost:3000/chat").build()?;

    event_loop.run(move |event, _, control_flow| {
        *control_flow = ControlFlow::Wait;

        if let Event::WindowEvent {
            event: WindowEvent::CloseRequested,
            ..
        } = event
        {
            *control_flow = ControlFlow::Exit
        }
    });
}

#[cfg(feature = "ssr")]
#[tokio::main]
async fn main() -> anyhow::Result<()> {
    use std::future::IntoFuture;

    use leptos::*;
    use llama_forge_rs::server::app;
    use tracing;

    setup_tracing();

    tracing::debug!("Hello from async_main");

    let conf = get_configuration(None).await.unwrap();
    // TODO Add figment + app Config struct and convert into leptos /tokio conf
    let leptos_options = conf.leptos_options;
    let addr = leptos_options.site_addr;

    // TODO Pass through endpoint ip address of conf to llama.cpp & make llama.cpp port configurable
    tracing::info!(
        "[MOP] listening on http://{} :{} (app) :{} (llama.cpp)",
        &addr.ip(),
        &addr.port(),
        8080
    );
    let listener = tokio::net::TcpListener::bind(&addr).await.unwrap();

    let serve = axum::serve(listener, app(leptos_options).await.into_make_service()).into_future();

    let proxy_man_fut = async move {
        use llama_proxy_man::{config::AppConfig, start_server};
        let config = AppConfig::default_figment();
        start_server(config).await?;

        Ok::<(), anyhow::Error>(())
    };

    let (serve_res, backend_res, wry_res) = tokio::try_join!(
        tokio::spawn(serve),
        tokio::task::spawn_blocking(wry_main),
        tokio::spawn(proxy_man_fut),
    )?;

    serve_res?;
    backend_res?;
    wry_res?;

    Ok(())
}

#[cfg(not(feature = "ssr"))]
pub fn main() {
    // no client-side main function
    // unless we want this to work with e.g., Trunk for a purely client-side app
    // see lib.rs for hydration function instead
}
